from pytest import mark

from classipypi import list_tags
from classipypi.interfaces import ListingConfig


@mark.parametrize(
    "exclude,expected",
    [
        (
            [],
            [
                "Development Status :: 1 - Planning",
                "Development Status :: 2 - Pre-Alpha",
                "Development Status :: 3 - Alpha",
                "Development Status :: 4 - Beta",
                "Development Status :: 5 - Production/Stable",
                "Development Status :: 6 - Mature",
                "Development Status :: 7 - Inactive",
            ],
        ),
        (
            ["Alpha"],
            [
                "Development Status :: 1 - Planning",
                "Development Status :: 4 - Beta",
                "Development Status :: 5 - Production/Stable",
                "Development Status :: 6 - Mature",
                "Development Status :: 7 - Inactive",
            ],
        ),
    ],
)
def test_list_dev_statuses(exclude, expected):
    config = ListingConfig(include=["Development Status"], exclude=exclude)
    tags = list_tags(config)
    assert tags == expected


@mark.parametrize(
    "case_insensitive,search_query,expected",
    [
        (
            False,
            "search",
            [
                "Intended Audience :: Science/Research",
            ],
        ),
        (
            True,
            "search",
            [
                "Intended Audience :: Science/Research",
                "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
            ],
        ),
        (False, "SeArCh", []),
        (
            True,
            "SeArCh",
            [
                "Intended Audience :: Science/Research",
                "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
            ],
        ),
    ],
)
def test_list_tags_case_insensitivity_standard(
    case_insensitive,
    search_query,
    expected,
):
    config = ListingConfig(include=[search_query], case_insensitive=case_insensitive)
    tags = list_tags(config)
    assert tags == expected


toml_example_single = """classifiers = [
    "Intended Audience :: Science/Research",
]"""

toml_example_double = """classifiers = [
    "Intended Audience :: Science/Research",
    "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
]"""


@mark.parametrize(
    "case_insensitive,search_query,expected",
    [
        (False, "search", [toml_example_single]),
        (True, "search", [toml_example_double]),
    ],
)
def test_list_tags_case_insensitivity_toml(case_insensitive, search_query, expected):
    config = ListingConfig(
        include=[search_query],
        case_insensitive=case_insensitive,
        toml=True,
    )
    tags = list_tags(config)
    assert tags == expected


grouped_example_single = """Topic
  Text Processing
    Markup
      HTML"""

grouped_example_multi = """Topic
  Text Processing
    Markup
      HTML
      LaTeX
      Markdown
      SGML
      VRML
      XML
      reStructuredText"""


@mark.parametrize(
    "case_insensitive,search_query,expected",
    [
        (False, "Markup :: H", [grouped_example_single]),
        (True, "Markup", [grouped_example_multi]),
    ],
)
def test_list_tags_output(case_insensitive, search_query, expected):
    config = ListingConfig(include=[search_query], group=True)
    tags = list_tags(config)
    assert tags == expected



from pathlib import Path

from pydantic import BaseModel, model_validator

from .display import DisplayConfig

__all__ = ["SourceConfig", "SelectorConfig"]


class SourceConfig(BaseModel):
    query: str | None = None
    source: Path | None = None

    @model_validator(mode="after")
    @classmethod
    def mutually_exclusive_and_required(cls, self):
        if self.query and self.source:
            raise ValueError("Cannot provide both 'query' and 'source'.")
        if not (self.query or self.source):
            raise ValueError("Must provide one of 'query' and 'source'.")
        return self


class SelectorConfig(DisplayConfig, SourceConfig):
    """
    Configure input source and output display.

      :param query: The query string.
      :param source: The source code.
      :param toml: Whether to display the tags as a TOML-compatible list.
      :param group: Whether to display tags grouped by section.
    """



from pydantic import BaseModel

from .display import DisplayConfig

__all__ = ["FilterConfig", "ListingConfig"]


class FilterConfig(BaseModel):
    include: list[str] = []
    exclude: list[str] = []
    case_insensitive: bool = False


class ListingConfig(DisplayConfig, FilterConfig):
    """
    Configure input filtering and output display.

      :param include: Strings to filter tags for.
      :param exclude: Strings to filter tags against.
      :param case_insensitive: Whether to match case-insensitively.
      :param toml: Whether to display the tags as a TOML-compatible list.
      :param group: Whether to display tags grouped by section.
    """



from sys import stderr
from textwrap import indent

import defopt
from pydantic import ValidationError

from .core import list_tags, select_tags
from .interfaces import ListingConfig, SelectorConfig

__all__ = ["run_cli"]

commands = {"sel": SelectorConfig, "ls": ListingConfig}


def handle_validation_error(ve: ValidationError) -> str:
    chosen_command = {v.__name__: k for k, v in commands.items()}[ve.title]
    error_msgs = "\n".join(str(e["ctx"]["error"]) for e in ve.errors())
    msg = f"Invalid {chosen_command!r} command:\n" + indent(error_msgs, prefix="- ")
    print(msg, end="\n\n", file=stderr)
    return chosen_command


def run_cli():
    try:
        config = defopt.run(commands, no_negated_flags=True)
    except ValidationError as ve:
        chosen_command = handle_validation_error(ve)
        try:
            defopt.run(commands[chosen_command], argv=["-h"], no_negated_flags=True)
        except SystemExit as exc:
            exc.code = 1
            raise
    else:
        match config:
            case SelectorConfig():
                print(f"Got {config=}")
                result = select_tags(config)
            case ListingConfig():
                result = list_tags(config)
        print("\n".join(result))



from collections import defaultdict

__all__ = ["build_nested_dict", "nest_tags"]


def build_nested_dict():
    """
    Helper function to create a nested defaultdict.
    """
    return defaultdict(build_nested_dict)


def nest_tags(tags: list[str]) -> str:
    """
    Organize tags into a nested dictionary structure.
    """
    nested_tags = build_nested_dict()
    for tag in tags:
        path = tag.split(" :: ")
        current_level = nested_tags
        for part in path:
            current_level = current_level[part]

    def format_nested_tags(nested_dict, level=0):
        """
        Function to format the nested tags as a string
        """
        result = ""
        for key, sub_dict in nested_dict.items():
            prefix = " " * 2 * level  # 2 spaces per indentation level
            result += f"{prefix}{key}\n"
            if isinstance(sub_dict, defaultdict):
                result += format_nested_tags(sub_dict, level + 1)
        return result

    return format_nested_tags(nested_tags).rstrip("\n")



import tomli_w
from trove_classifiers import sorted_classifiers

from ..interfaces import ListingConfig
from .tag_trees import nest_tags

__all__ = ["list_tags"]


def list_tags(config: ListingConfig) -> list[str]:
    tags = sorted_classifiers
    # Filtering
    if config.include:
        tags = [
            tag
            for tag in tags
            for query in config.include
            if (
                (query.lower() in tag.lower())
                if config.case_insensitive
                else (query in tag)
            )
        ]
    if config.exclude:
        tags = [
            tag
            for tag in tags
            for query in config.exclude
            if (
                (query.lower() not in tag.lower())
                if config.case_insensitive
                else (query not in tag)
            )
        ]
    # Presentation
    if config.group:
        tags = [nest_tags(tags)]
    elif config.toml:
        tags = [tomli_w.dumps({"classifiers": tags}).strip()]
    return tags
